
//		Acquire.cpp - internet file download plugin for TCC


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <WinInet.h>


#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					0
#define VER_BUILD					4
#define VER_PATCH					2

#define PLUGIN_NAME					L"Acquire"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			32768

#define QUOTE						0x22
#define APOS						0x27
#define GRAVE						0x60

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )



const wchar_t PluginFeatures[]		= L"acquire,_acqagent,_acqfile,_acqsize,@ifilesize,@ifiletime2";


const wchar_t HelpWindowTitle[]		= L"Acquire Plugin Help";


const wchar_t AcquireHelpText[]		= L"Download a file from the internet.\n\
\n\
ACQUIRE /G /N /O /Q url localname \r\n\
\n\
\t/G   show progress\n\
\t/N   not really\n\
\t/O   overwrite local file\n\
\t/Q   quiet\n\
\n\
The URL is required; it must begin with HTTP:// or FTP://\n\
The localname is optional.  If localname is a single dash,  \n\
or if it is not given and output is redirected, the file\n\
will be dumped to standard output.  If localname is not\n\
specified and output isn\u2019t redirected, or if localname is\n\
a directory, Acquire will generate a filename based on\n\
the URL.\n";


const wchar_t _AcqAgentHelpText[] = L"_ACQAGENT \r\n\
\n\
Returns the user agent string generated for use by   \n\
this plugin.\n";


const wchar_t _AcqFileHelpText[] = L"_ACQFILE \r\n\
\n\
Returns the filename of the last local file created by   \n\
ACQUIRE, or *ERROR* if the last transfer failed.\n";


const wchar_t _AcqSizeHelpText[] = L"_ACQSIZE \r\n\
\n\
Returns the size in bytes of the last local file created   \n\
by ACQUIRE, or -1 if the last transfer failed.\n";


const wchar_t F_IFileSizeHelpText[]		= L"@IFILESIZE[url] \r\n\
\n\
Returns the size, in bytes, of the file named by url,   \n\
or -1 on any error.\n";


const wchar_t F_IFileTime2HelpText[]	= L"@IFILETIME2[url,fmt] \r\n\
\n\
Returns the time stamp of the file named by url,   \n\
or -1 on any error.\n";



const wchar_t StdoutFilename[]		= L"-";
const wchar_t DefaultFilename[]		= L"index.html";
const wchar_t ErrorString[]			= L"*ERROR*";
const wchar_t EraseLine[]			= L"\r                                                            \r";


const wchar_t Windows1252[0x20] = { 0x20ac, 0x0000, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021, 0x02c6, 0x2030, 0x0160, 0x2039, 0x0152, 0x0000, 0x017d, 0x0000,
									0x0000, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014, 0x02dc, 0x2122, 0x0161, 0x203a, 0x0153, 0x0000, 0x017e, 0x0178 };


wchar_t LastFilename[MAX_FILENAME_LEN]		= L"";
unsigned __int64 LastFileBytesWritten		= 0;

wchar_t AgentString[SHORT_BUF_LEN]			= L"";

OSVERSIONINFOEXW OSVerInfo;



#include "DateToString.cpp"



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}


#include "ParseArgs.cpp"


BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



#ifdef _WIN64


BOOL IsWin64()
{
	return TRUE;
}

#else	//	_WIN64


BOOL IsWin64()
{
	wchar_t Wow64Dir[MAX_PATH]		= L"";


	GetSystemWow64Directory( Wow64Dir, MAX_PATH );


	if ( Wow64Dir[0] == '\0' )
		return FALSE;

	if ( !QueryIsDirectory( Wow64Dir ) )
		return FALSE;

	return TRUE;
}


#endif	//	_WIN64



int InitializeAgentString()
{
	wchar_t varbuf[SHORT_BUF_LEN]		= L" %_cmdproc/%_4ver (%@if[%_x64 == 1,x64,x86])";

	wsprintf( AgentString, L"%s/%u.%u.%u", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );
	ExpandVariables( varbuf, 1 );
	wcscat_s( AgentString, SHORT_BUF_LEN, varbuf );

	wsprintf( AgentString + wcslen( AgentString ), L" Windows/%u.%u.%u (%s)", OSVerInfo.dwMajorVersion, OSVerInfo.dwMinorVersion, OSVerInfo.dwBuildNumber, IsWin64() ? L"x64" : L"x86" );
	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/acquire.html";
	piInfo.pszDescription	= (LPTSTR) L"Downloads files from the internet.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	GetVersion2( &OSVerInfo );
	InitializeAgentString();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int ParseInt( LPCTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	
	DWORD FM_Flags					= FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK;

	static HMODULE WinINetHandle	= NULL;
	static HMODULE WinHttpHandle	= NULL;


	if ( ( err >= 12000 ) && ( err <= 12199 ) ) {
		if ( !WinINetHandle )
			WinINetHandle = GetModuleHandle( L"wininet.dll" );

		if ( WinINetHandle )
			FM_Flags |= FORMAT_MESSAGE_FROM_HMODULE;
	}


	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FM_Flags, WinINetHandle, err, 0, outString, SHORT_BUF_LEN, NULL ) ) {
		wsprintf( outString, L"Error %d", err );

		if ( ( err >= 12000 ) && ( err <= 12199 ) ) {
			if ( !WinHttpHandle )
				WinHttpHandle = GetModuleHandle( L"winhttp.dll" );

			if ( WinHttpHandle ) {
				FM_Flags |= FORMAT_MESSAGE_FROM_HMODULE;

				if ( !FormatMessage( FM_Flags, WinHttpHandle, err, 0, outString, SHORT_BUF_LEN, NULL ) )
					wsprintf( outString, L"Error %d", err );
			}
		}
	}

#ifdef _WININET_

	if ( err == ERROR_INTERNET_EXTENDED_ERROR ) {
		wchar_t ServerError[SHORT_BUF_LEN]		= L"";
		DWORD ErrorVal							= 0;
		DWORD BufLen							= SHORT_BUF_LEN;

		if ( InternetGetLastResponseInfo( &ErrorVal, ServerError, &BufLen ) ) {
			if ( ServerError[0] ) {
				wcscat_s( outString, SHORT_BUF_LEN, L" : " );
				wcscat_s( outString, SHORT_BUF_LEN, ServerError );
			}
		}
	}

#endif	//	_WININET_

	size_t l = wcslen( outString );

	if ( l > 0 ) {
		while ( l && iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}
	}

	return 0;
}



int GetHttpErrorText( int HttpError, LPTSTR outString, size_t buflen )
{
	if ( !outString || ( buflen < 64 ) )
		return -1;


	if ( HttpError < 300 ) {
		outString[0] = '\0';
		return 0;
	}


	switch ( HttpError ) {

	case 300 :
		wcscpy_s( outString, buflen, L"300 - Moved Permanently" );
		break;

	case 400 :
		wcscpy_s( outString, buflen, L"400 - Bad Request" );
		break;

	case 401 :
		wcscpy_s( outString, buflen, L"401 - Unauthorized" );
		break;

	case 402 :
		wcscpy_s( outString, buflen, L"402 - Payment Required" );
		break;

	case 403 :
		wcscpy_s( outString, buflen, L"403 - Forbidden" );
		break;

	case 404 :
		wcscpy_s( outString, buflen, L"404 - Not Found" );
		break;

	case 405 :
		wcscpy_s( outString, buflen, L"405 - Method Not Allowed" );
		break;

	case 406 :
		wcscpy_s( outString, buflen, L"406 - Not Acceptable" );
		break;

	case 407 :
		wcscpy_s( outString, buflen, L"407 - Proxy Authentication Required" );
		break;

	case 408 :
		wcscpy_s( outString, buflen, L"408 - Request Timeout" );
		break;

	case 410 :
		wcscpy_s( outString, buflen, L"410 - Gone" );
		break;

	case 414 :
		wcscpy_s( outString, buflen, L"414 - URI Too Long" );
		break;

	case 418 :
		wcscpy_s( outString, buflen, L"418 - I'm a teapot" );
		break;

	case 429 :
		wcscpy_s( outString, buflen, L"429 - Too Many Requests" );
		break;

	case 431 :
		wcscpy_s( outString, buflen, L"431 - Request Header Fields Too Largs" );
		break;

	case 451 :
		wcscpy_s( outString, buflen, L"451 - Unavailable For Legal Reasons" );
		break;

	case 500 :
		wcscpy_s( outString, buflen, L"500 - Internal server error" );
		break;

	case 501 :
		wcscpy_s( outString, buflen, L"501 - Not Implemented" );
		break;

	case 502 :
		wcscpy_s( outString, buflen, L"502 - Bad Gateway" );
		break;

	case 503 :
		wcscpy_s( outString, buflen, L"503 - Service unavailable" );
		break;

	case 504 :
		wcscpy_s( outString, buflen, L"504 - Gateway Timeout" );
		break;

	case 505 :
		wcscpy_s( outString, buflen, L"505 - HTTP Version Not Supported" );
		break;

	default :
		wsprintf( outString, L"%u - Other HTTP Error", HttpError );
	}

	return 0;
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



LPTSTR FindFilenameStart( LPTSTR Filename )
{
	if ( !Filename )
		return NULL;


	LPTSTR Start		= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == ':' ) || ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) )
			Start = NULL;
		else if ( !Start )
			Start = Filename + i;
	}

	return Start;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



int Type_Command( LPTSTR Args )
{
	if ( !Args )
		return -1;


	wchar_t CommandBuf[MAX_ARG_LEN] = L"*type ";
	wcscat_s( CommandBuf, MAX_ARG_LEN, Args );
	return Command( CommandBuf, 0 );
}



unsigned __int64 GetFtpFileSize( HINTERNET Connect, LPTSTR Url )
{
	if ( !Connect || !Url )
		return -1;

	if ( Url[0] == '\0' )
		return -1;


	HINTERNET FtpConnection		= NULL;
	HINTERNET FtpFileHandle		= NULL;
	DWORD FileSizeLow			= 0;
	DWORD FileSizeHigh			= 0;
	int LastError				= 0;

	URL_COMPONENTS Components;

	wchar_t Scheme[SHORT_BUF_LEN]			= L"";
	wchar_t HostName[MAX_ARG_LEN]			= L"";
	wchar_t UrlPath[MAX_FILENAME_LEN]		= L"";
	wchar_t ExtraInfo[MAX_ARG_LEN]			= L"";
	wchar_t UserName[SHORT_BUF_LEN]			= L"";
	wchar_t Password[SHORT_BUF_LEN]			= L"";


	memset( &Components, 0, sizeof( URL_COMPONENTS ) );
	Components.dwStructSize			= sizeof( URL_COMPONENTS );
	Components.lpszScheme			= Scheme;
	Components.dwSchemeLength		= SHORT_BUF_LEN;
	Components.lpszHostName			= HostName;
	Components.dwHostNameLength		= MAX_ARG_LEN;
	Components.lpszUrlPath			= UrlPath;
	Components.dwUrlPathLength		= MAX_FILENAME_LEN;
	Components.lpszExtraInfo		= ExtraInfo;
	Components.dwExtraInfoLength	= MAX_ARG_LEN;
	Components.lpszUserName			= UserName;
	Components.dwUserNameLength		= SHORT_BUF_LEN;
	Components.lpszPassword			= Password;
	Components.dwPasswordLength		= SHORT_BUF_LEN;

	InternetCrackUrl( Url, 0, 0, &Components );


	FtpConnection = InternetConnect( Connect, HostName, Components.nPort, UserName[0] ? UserName : NULL, ( UserName[0] && Password[0] ) ? Password : NULL, INTERNET_SERVICE_FTP, INTERNET_FLAG_PASSIVE, NULL );

	if ( FtpConnection ) {
		FtpFileHandle = FtpOpenFile( FtpConnection, UrlPath, GENERIC_READ, FTP_TRANSFER_TYPE_BINARY | INTERNET_FLAG_RELOAD, NULL );

		if ( FtpFileHandle ) {
			FileSizeLow = FtpGetFileSize( FtpFileHandle, &FileSizeHigh );
			LastError = GetLastError();

			InternetCloseHandle( FtpFileHandle );
		}

		InternetCloseHandle( FtpConnection );
	}

	if ( LastError )
		return -1;

	//  wprintf( L"   * FileSize = %I64u   GetLastError() = %d\r\n", ( (unsigned __int64) FileSizeHigh << 32 ) | FileSizeLow, LastError );
	return ( (unsigned __int64) FileSizeHigh << 32 ) | FileSizeLow;
}



unsigned __int64 GetHttpFileSize( HINTERNET RemoteFileHandle )
{
	if ( !RemoteFileHandle )
		return -1;


	DWORD FileSize		= 0;
	DWORD BufLen		= sizeof( DWORD );


	if ( !HttpQueryInfoW( RemoteFileHandle, HTTP_QUERY_CONTENT_LENGTH | HTTP_QUERY_FLAG_NUMBER, &FileSize, &BufLen, 0 ) )
		return -1;

	return FileSize;
}



int GetFtpFileTime( HINTERNET Connect, LPTSTR Url, SYSTEMTIME *TimeStamp, BOOL LocalTime )
{
	if ( !Connect || !Url || !TimeStamp )
		return 0;

	if ( Url[0] == '\0' )
		return 0;


	HINTERNET FtpConnection		= NULL;
	HINTERNET SearchHandle		= NULL;
	WIN32_FIND_DATA Found;
	BOOL Success				= FALSE;

	URL_COMPONENTS Components;

	wchar_t Scheme[SHORT_BUF_LEN]			= L"";
	wchar_t HostName[MAX_ARG_LEN]			= L"";
	wchar_t UrlPath[MAX_FILENAME_LEN]		= L"";
	wchar_t ExtraInfo[MAX_ARG_LEN]			= L"";
	wchar_t UserName[SHORT_BUF_LEN]			= L"";
	wchar_t Password[SHORT_BUF_LEN]			= L"";


	memset( &Components, 0, sizeof( URL_COMPONENTS ) );
	Components.dwStructSize			= sizeof( URL_COMPONENTS );
	Components.lpszScheme			= Scheme;
	Components.dwSchemeLength		= SHORT_BUF_LEN;
	Components.lpszHostName			= HostName;
	Components.dwHostNameLength		= MAX_ARG_LEN;
	Components.lpszUrlPath			= UrlPath;
	Components.dwUrlPathLength		= MAX_FILENAME_LEN;
	Components.lpszExtraInfo		= ExtraInfo;
	Components.dwExtraInfoLength	= MAX_ARG_LEN;
	Components.lpszUserName			= UserName;
	Components.dwUserNameLength		= SHORT_BUF_LEN;
	Components.lpszPassword			= Password;
	Components.dwPasswordLength		= SHORT_BUF_LEN;

	InternetCrackUrl( Url, 0, 0, &Components );


	if ( ( UrlPath[0] == '\0' ) || !wcscmp( UrlPath, L"/" ) )
		return 0;


	FtpConnection = InternetConnect( Connect, HostName, Components.nPort, UserName[0] ? UserName : NULL, ( UserName[0] && Password[0] ) ? Password : NULL, INTERNET_SERVICE_FTP, INTERNET_FLAG_PASSIVE, NULL );

	if ( FtpConnection ) {
		SearchHandle = FtpFindFirstFileW( FtpConnection, UrlPath, &Found, 0, NULL );

		if ( SearchHandle ) {
			Success = TRUE;
			InternetCloseHandle( SearchHandle );
		}

		InternetCloseHandle( FtpConnection );
	}

	if ( Success ) {
		if ( LocalTime )
			FileTimeToSystemTime( &( Found.ftLastWriteTime ), TimeStamp );
		else {
			SYSTEMTIME Temp;

			FileTimeToSystemTime( &( Found.ftLastWriteTime ), &Temp );
			TzSpecificLocalTimeToSystemTime( NULL, &Temp, TimeStamp );
		}

		return 1;
	}

	return 0;
}



DWORD GetHttpStatus( HINTERNET InternetHandle )
{
	DWORD Status = 0;
	DWORD BufferSize = sizeof( Status );

	HttpQueryInfo( InternetHandle, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &Status, &BufferSize, 0 );

	if ( Status >= 400 )
		return Status;
	else
		return 0;
}



int GetHttpFileTime( HINTERNET Connect, LPTSTR Url, SYSTEMTIME *TimeStamp, BOOL LocalTime )
{
	if ( !Connect || !Url || !TimeStamp )
		return 0;


	HINTERNET RemoteFileHandle		= NULL;
	DWORD OpenUrlFlags				= INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_RELOAD | INTERNET_FLAG_PASSIVE;
	DWORD HttpStatus				= 0;
	SYSTEMTIME FileTime;
	DWORD BufLen					= sizeof( SYSTEMTIME );
	int rv							= 0;


	RemoteFileHandle = InternetOpenUrl( Connect, Url, NULL, 0, OpenUrlFlags, NULL );

	if ( RemoteFileHandle ) {
		HttpStatus = GetHttpStatus( RemoteFileHandle );

		if ( HttpStatus >= 300 )
			rv = 0;
		else if ( !HttpQueryInfoW( RemoteFileHandle, HTTP_QUERY_LAST_MODIFIED | HTTP_QUERY_FLAG_SYSTEMTIME, &FileTime, &BufLen, 0 ) )
			rv = 0;
		else {
			if ( LocalTime )
				SystemTimeToTzSpecificLocalTime( NULL, &FileTime, TimeStamp );
			else
				*TimeStamp = FileTime;

			rv = 1;
		}

		InternetCloseHandle( RemoteFileHandle );
	}

	return rv;
}



int DisplayTicksAsHMS( unsigned __int64 Ticks )
{

	unsigned __int64 Hours		= 0;
	unsigned __int64 Minutes	= 0;
	unsigned __int64 Seconds	= 0;

	Seconds = Ticks / 1000;
	Minutes = Seconds / 60;
	Seconds %= 60;
	Hours = Minutes / 60;
	Minutes %= 60;

	_cwprintf( L"%2I64u:%02I64u:%02I64u", Hours, Minutes, Seconds );
	return 0;
}



int DisplaySizeCosmetic( unsigned __int64 Size )
{
	if ( Size >= ( (unsigned __int64) 1 << 40 ) )
		_cwprintf( L"%7.1f TB", (double) Size / (double) ( (unsigned __int64) 1 << 40 ) );
	else if ( Size >= ( (unsigned __int64) 1 << 30 ) )
		_cwprintf( L"%7.1f GB", (double) Size / (double) ( (unsigned __int64) 1 << 30 ) );
	else if ( Size >= ( 1 << 20 ) )
		_cwprintf( L"%7.1f MB", (double) Size / (double) ( 1 << 20 ) );
	else
		_cwprintf( L"%7.1f KB", (double) Size / (double) ( 1 << 10 ) );

	return 0;
}



int ConsoleHandle( HANDLE inHandle )
{
	if ( !inHandle )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	if ( !GetConsoleMode( inHandle, &Mode ) )
		return 0;

	if ( Mode == UINT_MAX )
		return 0;
	else
		return 1;
}



int ConsoleHandle( DWORD inHandle )
{
	return ConsoleHandle( (HANDLE) inHandle );
}



/*
int MakeUnsafe( LPTSTR String )
{
	if ( !String )
		return -1;

	int k = 0;

	for ( size_t i = 0; String[i]; i++ ) {
		if ( ( String[i] >= 0xff01 ) && ( String[i] <= 0xff5e ) ) {
			String[i] -= 0xfee0;
			k++;
		}
	}

	return k;
}
*/



int GoodProtocol( LPTSTR inString )
{
	if ( !inString )
		return 0;

	if ( !_wcsnicmp( inString, L"ftp://", 6 ) )
		return 2;

	if ( !_wcsnicmp( inString, L"http://", 7 ) )
		return 1;

	if ( !_wcsnicmp( inString, L"https://", 8 ) )
		return 1;

	return 0;
}



int AppendBackslash( LPTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t ch = '\0';


	for ( size_t i = 0; Filename[i]; i++ )
		ch = Filename[i];

	if ( ( ch != '/' ) && ( ch != '\\' ) )
		wcscat_s( Filename, MAX_FILENAME_LEN, L"\\" );

	return 0;
}



int FixFilenameCharacters( LPTSTR inFilename, LPTSTR outFilename )
{
	if ( !inFilename || !outFilename )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";

	size_t j		= 0;
	wchar_t ch		= '\0';


	for ( size_t i = 0; inFilename[i]; i++ ) {
		ch = inFilename[i];

		if ( ch == '%' ) {
			if ( iswxdigit( inFilename[i + 1] ) ) {
				if ( iswxdigit( inFilename[i + 2] ) ) {

					if ( iswdigit( inFilename[i + 1] ) )
						ch = 0x10 * ( inFilename[i + 1] - '0' );
					else
						ch = 0x10 * ( towupper( inFilename[i + 1] ) - 'A' + 10 );

					if ( iswdigit( inFilename[i + 2] ) )
						ch |= ( inFilename[i + 2] - '0' );
					else
						ch |= ( towupper( inFilename[i + 2] ) - 'A' + 10 );

					i += 2;
				}
			}
		}

		if ( ( ch >= 0x80 ) && ( ch <= 0x9f ) )
			ch = Windows1252[ch - 0x80];

		if ( ( ch == QUOTE ) || ( ch == GRAVE ) )
			ch = APOS;
		else if ( ch == '<' )
			ch = 0x2039;
		else if ( ch == '>' )
			ch = 0x203a;
		else if ( ch == '|' )
			ch = '!';
		else if ( ( ch == '/' ) || ( ch == '\\' ) || ( ch == '%' ) || ( ch == ':' ) || ( ch == ';' ) || ( ch == '?' ) || ( ch == '*' ) )
			ch = '_';
		else if ( iswcntrl( ch ) )
			ch = '\0';

		if ( ch && ( j < ( MAX_FILENAME_LEN - 1 ) ) ) {
			temp[j++] = ch;
			temp[j] = '\0';
		}
	}

	wcscpy_s( outFilename, MAX_FILENAME_LEN, temp );
	return 0;
}



int CreateLocalFilename( LPTSTR SourceUrl, LPTSTR LocalFile )
{
	if ( !SourceUrl || !LocalFile )
		return -1;

	if ( SourceUrl[0] == '\0' )
		return -1;


	URL_COMPONENTS Components;

	wchar_t Scheme[SHORT_BUF_LEN]			= L"";
	wchar_t HostName[MAX_ARG_LEN]			= L"";
	wchar_t UrlPath[MAX_FILENAME_LEN]		= L"";
	wchar_t ExtraInfo[MAX_ARG_LEN]			= L"";
	wchar_t TempName[MAX_FILENAME_LEN]		= L"";

	memset( &Components, 0, sizeof( URL_COMPONENTS ) );
	Components.dwStructSize			= sizeof( URL_COMPONENTS );
	Components.lpszScheme			= Scheme;
	Components.dwSchemeLength		= SHORT_BUF_LEN;
	Components.lpszHostName			= HostName;
	Components.dwHostNameLength		= MAX_ARG_LEN;
	Components.lpszUrlPath			= UrlPath;
	Components.dwUrlPathLength		= MAX_FILENAME_LEN;
	Components.lpszExtraInfo		= ExtraInfo;
	Components.dwExtraInfoLength	= MAX_ARG_LEN;

	InternetCrackUrl( SourceUrl, 0, 0, &Components );

	/*
	Printf( L"   * HostName   = '%s'\r\n", HostName );
	Printf( L"   * UrlPath    = '%s'\r\n", UrlPath );
	Printf( L"   * ExtraInfo  = '%s'\r\n", ExtraInfo );
	*/

	AppendBackslash( LocalFile );

	
	LPTSTR NamePart = FindFilenameStart( UrlPath );		//	FilenamePart( UrlPath, NULL );
	BOOL AnyName = FALSE;


	if ( NamePart ) {
		if ( NamePart[0] ) {
			//	Printf( L"   * NamePart = '%s'\r\n", NamePart );
			FixFilenameCharacters( NamePart, TempName );

			if ( TempName[0] ) {
				wcscat_s( LocalFile, MAX_FILENAME_LEN, TempName );
				AnyName = TRUE;
			}
		}
	}

	if ( !AnyName )
		wcscat_s( LocalFile, MAX_FILENAME_LEN, DefaultFilename );

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	if ( !HelpText )
		return -1;


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI acquire( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t LocalFile[MAX_FILENAME_LEN]		= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	LPTSTR SourceUrl						= NULL;
	BOOL Overwrite							= FALSE;
	BOOL Verbose							= FALSE;
	BOOL Quiet								= FALSE;
	BOOL DumpToStdout						= FALSE;
	BOOL TypeHex							= FALSE;
	BOOL TypeLineNumbers					= FALSE;
	BOOL NotReally							= FALSE;
	BOOL ShowProgress						= FALSE;
	DWORD OpenUrlFlags						= INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_RELOAD | INTERNET_FLAG_PASSIVE;
	int UrlProtocol							= 0;

	int ProgressCount						= 0;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( AcquireHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( AcquireHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/G" ) )
				ShowProgress = TRUE;
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				TypeLineNumbers = TRUE;
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				NotReally = TRUE;
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Overwrite = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Verbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				TypeHex = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !SourceUrl )
				SourceUrl = arg[i];
			else if ( LocalFile[0] == '\0' ) {
				if ( wcscmp( arg[i], StdoutFilename ) == 0 )
					DumpToStdout = TRUE;
				else {
					QueryTrueName( arg[i], LocalFile );

					if ( LocalFile[0] == '\0' ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
						return 1;
					}
				}
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}



	if ( !SourceUrl ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing source url\r\n" );
		return 1;
	}

	/*
	MakeUnsafe( SourceUrl );
	*/

	UrlProtocol = GoodProtocol( SourceUrl );

	if ( !UrlProtocol ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad or missing protocol: \"%s\"\r\n", SourceUrl );
		return 1;
	}


	if ( !ConsoleHandle( STD_OUTPUT_HANDLE ) && ( LocalFile[0] == '\0' ) )
		DumpToStdout = TRUE;

	if ( DumpToStdout ) {
		GetTempPath( MAX_FILENAME_LEN, LocalFile );				//  GetTempDirectory( LocalFile );		AppendBackslash( LocalFile );
		GetTempFileName( LocalFile, L"ACQ", 0, LocalFile );		//	UniqueFileName( LocalFile, L"ACQ" );
		Overwrite = Quiet = TRUE;
	}

	if ( LocalFile[0] ) {
		if ( QueryIsURL( LocalFile ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Not a local filename: \"%s\"\r\n", LocalFile );
			return 1;
		}
	}
	
	if ( LocalFile[0] == '\0' ) {
		wcscpy_s( LocalFile, MAX_FILENAME_LEN, L"%_cwds" );
		ExpandVariables( LocalFile, 1 );
	}

	if ( QueryIsDirectory( LocalFile ) )
		CreateLocalFilename( SourceUrl, LocalFile );
	

	const unsigned int BufferSize		= 8192;
	BYTE Buffer[BufferSize]				= { 0 };

	DWORD BytesRead						= 0;
	DWORD BytesWritten					= 0;

	HINTERNET Connection				= NULL;
	HINTERNET RemoteFileHandle			= NULL;
	HANDLE LocalFileHandle				= NULL;

	int ReadReturn						= 0;
	int WriteReturn						= 0;
	int ReadErrors						= 0;
	int WriteError						= 0;
	int rv								= 2;
	DWORD HttpStatus					= 0;

	unsigned __int64 TotalBytesWritten	= 0;
	unsigned __int64 ExpectedFileSize	= 0;

	unsigned __int64 StartTime			= 0;
	unsigned __int64 CurrentTime		= 0;
	unsigned __int64 TimeAdjust			= 0;
	BOOL Underflow						= FALSE;
	unsigned __int64 TimeElapsed		= 0;
	//	unsigned __int64 EstTotalTime		= 0;
	//	unsigned __int64 EstTimeRemaining	= 0;

	unsigned __int64 permyriad			= 0;



	if ( Verbose ) {
		Printf( L"Source URL: %s\r\n", SourceUrl );
		Printf( L"Local file: %s\r\n", LocalFile );
		Printf( L"Agent string: %s\r\n", AgentString );
	}

	if ( NotReally ) {
		Printf( L"Wrote 0 bytes to \"%s\".\r\n", LocalFile );
		return 0;
	}

	if ( !Overwrite ) {
		if ( IsAFile( LocalFile ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File exists: \"%s\"\r\n", LocalFile );
			wcscpy_s( LastFilename, MAX_FILENAME_LEN, ErrorString );
			LastFileBytesWritten = 0;
			return 2;
		}
	}


	Connection = InternetOpen( AgentString, INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0 );

	if ( Connection ) {

		RemoteFileHandle = InternetOpenUrl( Connection, SourceUrl, NULL, 0, OpenUrlFlags, NULL );

		if ( RemoteFileHandle ) {

			if ( UrlProtocol == 1 )
				HttpStatus = GetHttpStatus( RemoteFileHandle );

			if ( UrlProtocol == 1 ) {
				DWORD BufferLength = sizeof( DWORD );
				HttpQueryInfo( RemoteFileHandle, HTTP_QUERY_CONTENT_LENGTH | HTTP_QUERY_FLAG_NUMBER, &ExpectedFileSize, &BufferLength, 0 );
			}
			else if ( UrlProtocol == 2 )
				ExpectedFileSize = GetFtpFileSize( Connection, SourceUrl );

			if ( ExpectedFileSize == (unsigned __int64) -1 )
				ExpectedFileSize = 0;


			if ( HttpStatus >= 300 ) {
				GetHttpErrorText( HttpStatus, ErrorText, SHORT_BUF_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%s: \"%s\"\r\n", ErrorText, SourceUrl );
			}
			else {
				LocalFileHandle = CreateFile( LocalFile, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );

				if ( LocalFileHandle != INVALID_HANDLE_VALUE ) {

					StartTime = GetTickCount();

					do {
						BytesRead = 0;
						ReadReturn = InternetReadFile( RemoteFileHandle, Buffer, BufferSize, &BytesRead );

						if ( ReadReturn ) {
							BytesWritten = 0;
							WriteReturn = WriteFile( LocalFileHandle, Buffer, BytesRead, &BytesWritten, NULL );

							if ( WriteReturn ) {
								TotalBytesWritten += BytesWritten;

								if ( ShowProgress ) {
									ProgressCount = ( ProgressCount + 1 ) & 15;
									if ( ProgressCount == 0 ) {

										CurrentTime = GetTickCount();
										if ( CurrentTime < StartTime ) {
											if ( !Underflow )
												TimeAdjust += ( 1LL << 32 );

											Underflow = TRUE;
										}
										else
											Underflow = FALSE;

										CurrentTime |= TimeAdjust;
										TimeElapsed = CurrentTime - StartTime;

										_cputws( L"\r" );
										DisplaySizeCosmetic( TotalBytesWritten );
										_cputws( L"   " );

										if ( ExpectedFileSize ) {
											_cwprintf( L"%5.1f%%   ", TotalBytesWritten * 100.0 / ExpectedFileSize );
											permyriad = ( TotalBytesWritten * 10000 ) / ExpectedFileSize;
										}

										DisplayTicksAsHMS( TimeElapsed );
										_cputws( L" " );

										/*
										if ( ExpectedFileSize && ( permyriad >= 200 ) ) {
											EstTotalTime = ( TimeElapsed * 10000 ) / permyriad;
											EstTimeRemaining = EstTotalTime - TimeElapsed;
											_cputws( L"   (" );
											DisplayTicksAsHMS( EstTimeRemaining );
											_cputws( L") " );
										}
										*/
									}
								}
							}
							else {
								WriteError = GetLastError();
								if ( WriteError == 0 )
									WriteError = -1;
							}
						}
						else {
							if ( ShowProgress )
								_cputws( EraseLine );

							GetErrorText( GetLastError(), ErrorText );
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Error reading from URL \"%s\":\r\n    %s\r\n", SourceUrl, ErrorText );
							ReadErrors++;
						}

					} while ( ReadReturn && BytesRead );

					if ( ShowProgress )
						_cputws( EraseLine );

					if ( WriteError ) {
						GetErrorText( WriteError, ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Error writing to file: \"%s\":\r\n    %s\r\n", LocalFile, ErrorText );
					}

					CloseHandle( LocalFileHandle );

					if ( !WriteError && !ReadErrors ) {
						rv = 0;

						if ( !Quiet ) {
							wchar_t shortbuf[SHORT_BUF_LEN] = L"";
							wsprintf( shortbuf, L"%I64u", TotalBytesWritten );
							InsertCommas( shortbuf, SHORT_BUF_LEN );		//  AddCommas( shortbuf );
							Printf( L"Wrote %s byte%s to \"%s\".\r\n", shortbuf, ( TotalBytesWritten == 1 ) ? L"" : L"s", LocalFile );
						}
					}
				}
				else {
					GetErrorText( GetLastError(), ErrorText );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error creating file \"%s\":\r\n    %s\r\n", LocalFile, ErrorText );
				}
			}

			InternetCloseHandle( RemoteFileHandle );
		}
		else {
			int err = GetLastError();
			GetErrorText( err, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening URL \"%s\":\r\n    %s\r\n", SourceUrl, ErrorText );
		}

		InternetCloseHandle( Connection );
	}
	else {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening internet: %s\r\n", ErrorText );
	}


	if ( DumpToStdout ) {
		wchar_t CmdBuf[MAX_FILENAME_LEN + 20] = L" ";

		if ( TypeLineNumbers )
			wcscat_s( CmdBuf, MAX_FILENAME_LEN + 20, L"/L " );
		else if ( TypeHex )
			wcscat_s( CmdBuf, MAX_FILENAME_LEN + 20, L"/X " );

		wcscat_s( CmdBuf, MAX_FILENAME_LEN + 20, L"\"" );
		wcscat_s( CmdBuf, MAX_FILENAME_LEN + 20, LocalFile );
		wcscat_s( CmdBuf, MAX_FILENAME_LEN + 20, L"\"" );
		//  Printf( L"   * CmdBuf = %s\r\n", CmdBuf );
		
		Type_Command( CmdBuf );				//	Type_Cmd( CmdBuf );

		DeleteFile( LocalFile );
	}


	if ( rv ) {
		wcscpy_s( LastFilename, MAX_FILENAME_LEN, ErrorString );
		LastFileBytesWritten = 0;
	}
	else if ( DumpToStdout ) {
		wcscpy_s( LastFilename, MAX_FILENAME_LEN, StdoutFilename );
		LastFileBytesWritten = TotalBytesWritten;
	}
	else {
		wcscpy_s( LastFilename, MAX_FILENAME_LEN, LocalFile );
		LastFileBytesWritten = TotalBytesWritten;
	}

	return rv;
}



DLLExports int WINAPI _acqagent( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_ARG_LEN, AgentString );
	return 0;
}



DLLExports int WINAPI _acqfile( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_FILENAME_LEN, LastFilename );
	return 0;
}



DLLExports int WINAPI _acqsize( LPTSTR outString )
{
	if ( !outString )
		return 1;


	if ( !wcscmp( LastFilename, ErrorString ) )
		wcscpy_s( outString, MAX_ARG_LEN, L"-1" );
	else
		wsprintf( outString, L"%I64u", LastFileBytesWritten );

	return 0;
}



/*
DLLExports int WINAPI _acqver( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wsprintf( outString, L"%u.%u.%u (%s)", VER_MAJOR, VER_MINOR, VER_BUILD, PLATFORM );
	return 0;
}
*/



DLLExports int WINAPI f_ifilesize( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	int UrlProtocol				= 0;
	HINTERNET Connection		= NULL;
	HINTERNET RemoteFileHandle	= NULL;
	DWORD OpenUrlFlags			= INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_RELOAD | INTERNET_FLAG_PASSIVE;
	DWORD HttpStatus			= 0;

	unsigned __int64 FileSize	= 0;


	TrimLine( String );


	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	UrlProtocol = GoodProtocol( String );
	if ( !UrlProtocol ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad or missing protocol: \"%s\"\r\n", String );
		return -1;
	}


	Connection = InternetOpen( AgentString, INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0 );

	if ( Connection ) {
		RemoteFileHandle = InternetOpenUrl( Connection, String, NULL, 0, OpenUrlFlags, NULL );

		if ( RemoteFileHandle ) {
			if ( UrlProtocol == 1 )
				HttpStatus = GetHttpStatus( RemoteFileHandle );

			if ( HttpStatus >= 300 )
				FileSize = -1;
			else if ( UrlProtocol == 1 )
				FileSize = GetHttpFileSize( RemoteFileHandle );
			else if ( UrlProtocol == 2 )
				FileSize = GetFtpFileSize( Connection, String );

			InternetCloseHandle( RemoteFileHandle );
		}
		else
			FileSize = -1;

		InternetCloseHandle( Connection );
	}
	else {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening internet: %s\r\n", ErrorText );
		return -1;
	}


	if ( FileSize == (unsigned __int64) -1 )
		wcscpy_s( String, MAX_ARG_LEN, L"-1" );
	else
		wsprintf( String, L"%I64u", FileSize );

	return 0;
}



DLLExports int WINAPI f_ifiletime2( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	int UrlProtocol				= 0;
	HINTERNET Connection		= NULL;
	SYSTEMTIME TimeStamp;
	BOOL Success				= FALSE;
	unsigned int Fmt			= 4;
	BOOL LocalTime				= FALSE;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Fmt, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad format: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( Fmt & 1024 )
		LocalTime = TRUE;


	UrlProtocol = GoodProtocol( arg[0] );
	if ( !UrlProtocol ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad or missing protocol: \"%s\"\r\n", arg[0] );
		return -1;
	}


	Connection = InternetOpen( AgentString, INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0 );

	if ( Connection ) {

		if ( UrlProtocol == 1 )
			Success = GetHttpFileTime( Connection, arg[0], &TimeStamp, LocalTime );
		else if ( UrlProtocol == 2 )
			Success = GetFtpFileTime( Connection, arg[0], &TimeStamp, LocalTime );

		InternetCloseHandle( Connection );
	}
	else {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening internet: %s\r\n", ErrorText );
		return -1;
	}


	if ( Success ) {
		DateToString( String, TimeStamp.wYear, TimeStamp.wMonth, TimeStamp.wDay, Fmt );

		if ( Fmt & 2048 )
			wsprintf( String + wcslen( String ), L" %02u:%02u", TimeStamp.wHour, TimeStamp.wMinute );
		else
			wsprintf( String + wcslen( String ), L" %02u:%02u:%02u", TimeStamp.wHour, TimeStamp.wMinute, TimeStamp.wSecond );
	}
	else
		wcscpy_s( String, MAX_ARG_LEN, L"-1" );
	
	return 0;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	LPCTSTR HelpStr				= NULL;


	if ( !_wcsicmp( inString, L"acquire" ) )
		HelpStr = AcquireHelpText;
	else if ( !_wcsicmp( inString, L"_acqagent" ) )
		HelpStr = _AcqAgentHelpText;
	else if ( !_wcsicmp( inString, L"_acqfile" ) )
		HelpStr = _AcqFileHelpText;
	else if ( !_wcsicmp( inString, L"_acqsize" ) )
		HelpStr = _AcqSizeHelpText;
	else if ( !_wcsicmp( inString, L"@ifilesize" ) )
		HelpStr = F_IFileSizeHelpText;
	else if ( !_wcsicmp( inString, L"@ifiletime2" ) )
		HelpStr = F_IFileTime2HelpText;

	if ( HelpStr ) {
		MessageBox( GetForegroundWindow(), HelpStr, HelpWindowTitle, MB_OK );
		return 1;
	}
	
	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"acquire" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( AcquireHelpText ) );
		return 1;
	}
	else
		return 0;
}
